import math
from experiments.models import *
from lib.symbolic import Dist, Analyser, convert_equation, convert_expr, convert_to_bounds, z3_abs
from lib.interpreter import Evaluator, JaxEvaluator
from lib.utils import load_program, transform_program
from lib.typing import AssignmentStatement, BlockStatement, IfStatement, PrimType, Statement, WhileLoopProgram, WhileStatement
from typing import Dict, List, TypedDict, Tuple
from dataclasses import dataclass, field
import jax
import jax.numpy as jnp
from jax import grad, jit, vmap
from jax.nn import sigmoid, initializers, softplus, relu
import jax.random as jrandom
from sympy import Expr, Function, Integer, Symbol, Number, Add, Mul, Min, Max, Lambda, Pow, Piecewise, Lt, Abs, Le, Gt, Ge, Eq
from abc import ABC, abstractmethod
import optax
import string
import random
from pathlib import Path
from fractions import Fraction
import z3
from datetime import datetime
from pprint import pprint
from contextlib import redirect_stdout
import statistics
import os
from jax.scipy.special import logsumexp
import time

# SCALE = 1000
SCALE = 50 # Make SCALE_FACTOR = 1, for RepSM
SCALE_FACTOR = SCALE / 50
SHOULD_ROUND = False # Normally False, True for probfact
SHOULD_MAKE_SPARSE = True # For repulse_2d, RepSM
SAMPLE_COUNT = 10_000 # Normally 200, 1000 for faulty_loop_uniform
SAMPLE_DISTANCE = 1
REPSM_INIT = False

def make_sparse(params):
    new_params = dict(params)
    for k, v in new_params.items():
        if abs(v) < 0.05:
            new_params[k] = 0.0
    return new_params

def show(cand):
    return {k: float(v) for k, v in cand.items()}

class Dataset(TypedDict):
    pre: List[Tuple]
    asserting: List[Tuple]

class JaxDataset(TypedDict):
    pre: jnp.ndarray
    asserting: jnp.ndarray

class JaxAuxiliaryData(TypedDict):
    post: jnp.ndarray
    factors: jnp.ndarray

def softminus(input, scale=SCALE, smooth=True):
    if smooth:
        # return jnp.where(input < scale, input, scale + (input - scale) * 0.01)
        return -softplus(-input + scale) + scale
    else:
        return jnp.minimum(input, scale)

def round_2dp(params):
    new_params = dict(params)
    rounded_params = dict([(w, round((new_params[w]), 0)) for w in new_params])
    return rounded_params

class Loss(ABC):
    @abstractmethod
    def compute(self, penalty_safe_vector, penalty_unsafe_vector):
        pass

    @abstractmethod
    def log(self, config):
        pass

@dataclass
class ProblemData(ABC):
    pre_expr: Expr
    post_expr: Expr
    initial_state: jnp.ndarray
    keys: List[str]
    keys_other: List[str]
    keys_special: List[str]
    template_vars: List[Symbol]
    bias_vars: List[Symbol]
    safe_z3: z3.ExprRef
    unsafe_z3: z3.ExprRef
    inv_z3: z3.ExprRef
    non_increasing_cond_z3: z3.ExprRef
    indicating_cond_z3: z3.ExprRef
    prog: WhileLoopProgram
    dist: Dist

    @classmethod
    def create(cls, prog: WhileLoopProgram, dist: Dist, initial_state: Dict, **kwargs):
        keys = [var for var in dist.begin_map]
        keys_other = [f"{var}_{dist.begin_map[var]}" for var in keys]
        vars_in_body = cls.get_assigned(prog.body) - dist.begin_map.keys()
        keys_special = keys + [var for var in vars_in_body]
        real_keys = [f"{k}_{v}" for k, v in dist.begin_map.items() if dist.begin_type_map[k] == PrimType.REAL]
        initial_state = [initial_state[k] for k in keys]
        return cls(
            prog=prog,
            keys=keys,
            keys_other=keys_other,
            keys_special=keys_special,
            real_keys=real_keys,
            initial_state=jnp.asarray(initial_state),
            dist=dist,
            **kwargs
        )

    @classmethod
    def get_assigned(cls, statements: List[Statement]):
        vars = set()
        for stmt in statements:
            match stmt:
                case AssignmentStatement(name=name, expr=_):
                    vars.add(name)
                case IfStatement(branches=branches):
                    for (_, block) in branches:
                        vars |= cls.get_assigned(block.statements)
        return vars

    def get_cxex_count(self, params, total_safe_cxex, total_unsafe_cxex):
        safe_cxex, unsafe_cxex = self.get_safe_and_unsafe_cxex(params, total_safe_cxex, total_unsafe_cxex)
        len_safe_cxex = len(safe_cxex)
        len_unsafe_cxex = len(unsafe_cxex)
        if len_safe_cxex == total_safe_cxex:
            len_safe_cxex = math.inf
        if len_unsafe_cxex == total_unsafe_cxex:
            len_unsafe_cxex = math.inf
        return len_safe_cxex, len_unsafe_cxex

    def get_safe_and_unsafe_cxex(self, params, total_safe_cxex, total_unsafe_cxex):
        if SHOULD_MAKE_SPARSE:
            params = make_sparse(params)
        if SHOULD_ROUND:
            params = round_2dp(params)
        params_z3 = self.params_to_z3(params)
        safe_cxex = self.get_counterexamples(
            z3.And(self.safe_z3, self.inv_z3, params_z3), 
            self.non_increasing_cond_z3, 
            total=total_safe_cxex
        )    
        unsafe_cxex = self.get_counterexamples(
            z3.And(self.unsafe_z3, self.inv_z3, params_z3), 
            self.indicating_cond_z3, 
            total=total_unsafe_cxex
        )
        return safe_cxex, unsafe_cxex

    def params_to_z3(self, params):
        result = []
        for k, v in params.items():
            frac = Fraction(str(float(v)))
            result.append(z3.Real(k) == z3.Q(frac.numerator, frac.denominator))
        return z3.And(*result)
        
    def get_counterexamples(self, antecedent: z3.ExprRef, consequent: z3.ExprRef, total):
        def convert_to_dict(cxex):
            result = {}
            for k in self.keys_other:
                if k in self.real_keys:
                    result[k] = float(cxex[z3.Real(k)].as_fraction())
                else:
                    result[k] = cxex[z3.Int(k)].as_long()
            return result
        def convert_to_cond(cxex):
            result = []
            for k, v in cxex.items():
                if k in self.real_keys:
                    frac = Fraction(str(v))
                    var = z3.Real(k)
                    result.append(z3.Or(var <= frac - SAMPLE_DISTANCE, var >= frac + SAMPLE_DISTANCE))
                else:
                    result.append(z3.Int(k) != v)
            return z3.Or(*result)
        conds = [
            antecedent,
            z3.Not(consequent)
        ]
        cxex_list = []
        for i in range(0, total):
            solver = z3.Solver()
            solver.add(z3.And(*conds))
            result = solver.check()
            if result.r == -1:
                break
            cxex = convert_to_dict(solver.model())
            cxex_list.append(cxex)
            # print(convert_to_cond(cxex))
            conds.append(convert_to_cond(cxex))
        return cxex_list

@dataclass
class IsmProblemData(ProblemData):
    use_samples: bool
    real_keys: List[str] = field(default_factory=list)
    
    def is_really_valid(self, params, info):
        return self.get_cxex_count(params, 5, 5) == (0, 0)

    def get_real_bound(self, params, info):
        return float(self.get_initial_state_bound(params, info)) / SCALE

    def get_initial_state_bound(self, params, info):
        martingale = convert_jax(self.pre_expr, self.keys_other, smooth=False)
        return martingale(params, self.initial_state)

    def trace_loss(self, dataset: JaxDataset, loss: Loss, smooth=True):
        martingale = convert_jax(self.pre_expr, self.keys_other, smooth=smooth)

        if self.use_samples:
            get_fn = build_jax_auxiliary_dataset(dataset, self)
            aux_dataset = get_fn(38)
            # aux_dataset = get_fn(37)

            @jit
            def jax_loss(params, i: int):
                # aux_dataset = get_fn(i)
                # pre_matrix: N x V
                # batched_eta : params -> (_a x V) -> (_a x 1)
                batched_eta = vmap(martingale, in_axes=(None, 0))

                # eta_pre_matrix_reshaped : (N,)
                eta_pre_matrix_reshaped = batched_eta(params, dataset["pre"]).reshape((-1))

                # batched_batched_eta : params -> (_b x _a x V) -> (_b x _a x 1)
                batched_batched_eta = vmap(batched_eta, in_axes=(None, 0))
                eta_post_matrix = batched_batched_eta(params, aux_dataset["post"])
                eta_post_matrix_reshaped = eta_post_matrix.reshape(eta_post_matrix.shape[:2])   
                eta_post_matrix_reshaped_weighted = eta_post_matrix_reshaped * aux_dataset["factors"]     
                exp_eta_post_matrix = jnp.mean(eta_post_matrix_reshaped_weighted, axis=1)
                eta_asserting_matrix = batched_eta(params, dataset["asserting"]).reshape((-1))

                penalty_safe_vector = exp_eta_post_matrix - eta_pre_matrix_reshaped
                penalty_unsafe_vector = SCALE - eta_asserting_matrix
                # penalty_unsafe_vector = softminus(SCALE, smooth=self.smooth) - eta_asserting_matrix

                initial_state_bound = martingale(params, self.initial_state)
                return loss.compute(penalty_safe_vector, penalty_unsafe_vector, initial_state_bound)
        else:
            func = convert_jax(self.post_expr, self.keys_other, smooth=smooth)

            @jit
            def jax_loss(params, i: int):
                # pre_matrix: N x V
                # batched_eta : params -> (_a x V) -> (_a x 1)
                batched_eta = vmap(martingale, in_axes=(None, 0))

                # eta_pre_matrix_reshaped : (N,)
                eta_pre_matrix_reshaped = batched_eta(params, dataset["pre"]).reshape((-1))

                # batched_batched_eta : params -> (_b x _a x V) -> (_b x _a x 1)
                exp_eta_post_matrix = vmap(func, in_axes=(None, 0))(params, dataset["pre"])
                eta_asserting_matrix = batched_eta(params, dataset["asserting"]).reshape((-1))

                penalty_safe_vector = exp_eta_post_matrix - eta_pre_matrix_reshaped
                penalty_unsafe_vector = SCALE - eta_asserting_matrix
                # penalty_unsafe_vector = softminus(SCALE, smooth=self.smooth) - eta_asserting_matrix

                initial_state_bound = martingale(params, self.initial_state)
                return loss.compute(penalty_safe_vector, penalty_unsafe_vector, initial_state_bound)

        return jax_loss

    def get_helper_info(self, params, dataset):
        return {}

    def get_helper_info_other(self, params, dataset, aux_dataset):
        return {}

@dataclass
class RepsmProblemData(ProblemData):
    limit_expr: Expr
    use_samples: bool = False
    real_keys: List[str] = field(default_factory=list)
    epsilon = 1

    def get_real_bound(self, params, info):
        return float(self.get_initial_state_bound(params, info))

    def is_really_valid(self, params, info):
        if self.get_cxex_count(params, 5, 5) != (0, 0):
            return False
        template_vars_as_set = set(map(lambda x: x.name, self.template_vars))
        params_z3 = self.params_to_z3(params)
        initial = z3.Real("initial")
        result = z3.Real("result_0")
        initialCondition = initial == convert_expr(fix(self.pre_expr), self.dist, template_vars_as_set)
        if self.use_samples:
            limit = Fraction(str(float(info["limit"]))) * Fraction(101, 100)
            # print(limit)
        else:
            limit = convert_expr(self.limit_expr, self.dist, template_vars_as_set)
        limitCondition = z3_abs(initial - result) <= limit
        dist_with_result = self.dist.add_result_partial(fix(self.pre_expr))
        completeCondition = convert_to_bounds(dist_with_result.pdf, dist_with_result, template_vars_as_set)
        cxex = self.get_counterexamples(
            z3.And(params_z3, self.inv_z3, self.safe_z3, initialCondition, completeCondition), 
            limitCondition,
            5
        )
        # print("Cxex", cxex)
        return len(cxex) == 0

    def compute_direct(self, initial, eps, lim):
        lim = lim + 0.00001
        gamma = jnp.exp(- (eps ** 2) / (2 * (lim + eps) ** 2))
        alpha = jnp.exp((eps * initial) / (lim + eps) ** 2)
        final = alpha * ((gamma ** (jnp.abs(initial) / lim)) / (1 - gamma))
        return final

    def compute_direct_log(self, initial, eps, lim):
        logAlpha = (eps * initial) / (lim + eps) ** 2
        logGamma = - (eps ** 2) / (2 * (lim + eps) ** 2)
        log1MinusGamma = logsumexp(jnp.asarray([0, logGamma]), b=jnp.asarray([1, -1]))
        A = jnp.abs(initial) / lim
        return logAlpha + A * logGamma - log1MinusGamma

    def get_initial_state_bound(self, params, info):
        if self.use_samples:
            return self.get_initial_state_bound_with_limit(params, info["limit"])
        else:
            return self.get_initial_state_bound_with_limit_expr(params)

    def get_initial_state_bound_with_limit_expr(self, params):
        lim_func = convert_jax(self.limit_expr, [])
        limit = lim_func(params, None)
        return self.get_initial_state_bound_with_limit(params, limit)

    def get_initial_state_bound_with_limit(self, params, limit, smooth=False):
        martingale = convert_jax(self.pre_expr, self.keys_other, smooth=smooth)
        initial_state_bound = martingale(params, self.initial_state)
        return self.compute_direct(initial_state_bound, self.epsilon, limit)

    def get_helper_info(self, params, dataset):
        dataset = build_jax_dataset(dataset, self)
        aux_dataset = build_jax_auxiliary_dataset(dataset, self)(38)
        return self.get_helper_info_other(params, dataset, aux_dataset)

    def get_helper_info_other(self, params, dataset, aux_dataset):
        if self.use_samples:
            martingale = convert_jax(self.pre_expr, self.keys_other, smooth=False)

            batched_eta = vmap(martingale, in_axes=(None, 0))
            eta_pre_matrix_reshaped = batched_eta(params, dataset["pre"]).reshape((-1))
            eta_pre_matrix_reshaped2 = jnp.reshape(eta_pre_matrix_reshaped, (-1, 1))

            batched_batched_eta = vmap(batched_eta, in_axes=(None, 0))
            eta_post_matrix = batched_batched_eta(params, aux_dataset["post"])
            eta_post_matrix_reshaped = eta_post_matrix.reshape(eta_post_matrix.shape[:2])   
            eta_post_matrix_reshaped_weighted = eta_post_matrix_reshaped * aux_dataset["factors"]     
            exp_eta_post_matrix = jnp.mean(eta_post_matrix_reshaped_weighted, axis=1)
            eta_asserting_matrix = batched_eta(params, dataset["asserting"]).reshape((-1))
            max_diff = logsumexp(1000000 * jnp.abs(eta_post_matrix_reshaped - eta_pre_matrix_reshaped2)) / 1000000

            penalty_vector_safe = exp_eta_post_matrix - eta_pre_matrix_reshaped + self.epsilon
            penalty_vector_unsafe = -eta_asserting_matrix
            bound_approx = self.get_initial_state_bound_with_limit(params, max_diff)

            # print(eta_post_matrix_reshaped)
            # print((jnp.abs(eta_post_matrix_reshaped - eta_pre_matrix_reshaped2)))
            # print("Params", show(params))
            # print("Mean ReLU Safe", jnp.mean(relu(penalty_vector_safe)))
            # print("Mean ReLU Unsafe", jnp.mean(relu(penalty_vector_unsafe)))
            # print("Bound Approx", bound_approx)
            # if loss != None:
            #     print("Loss", loss.compute(penalty_vector_safe, penalty_vector_unsafe, bound_approx))

            return {"limit": jnp.max(jnp.abs(eta_post_matrix_reshaped - eta_pre_matrix_reshaped2))}
        else:
            return {}

    # def get_stat_info(self, dataset: JaxDataset):
    #     martingale = convert_jax(self.pre_expr, self.keys_other, smooth=True)
    #     get_fn = build_jax_auxiliary_dataset(dataset, self)
    #     aux_dataset = get_fn(38)

    #     def fn(params):
    #         batched_eta = vmap(martingale, in_axes=(None, 0))
    #         eta_pre_matrix_reshaped = batched_eta(params, dataset["pre"]).reshape((-1))
    #         eta_pre_matrix_reshaped2 = jnp.reshape(eta_pre_matrix_reshaped, (-1, 1))

    #         batched_batched_eta = vmap(batched_eta, in_axes=(None, 0))
    #         eta_post_matrix = batched_batched_eta(params, aux_dataset["post"])
    #         eta_post_matrix_reshaped = eta_post_matrix.reshape(eta_post_matrix.shape[:2])   
    #         eta_post_matrix_reshaped_weighted = eta_post_matrix_reshaped * aux_dataset["factors"]     
    #         exp_eta_post_matrix = jnp.mean(eta_post_matrix_reshaped_weighted, axis=1)
    #         a = jnp.abs(eta_post_matrix_reshaped - eta_pre_matrix_reshaped2)
    #         print(eta_post_matrix_reshaped)
    #         print(logsumexp(10000 * a) / 10000, jnp.max(a))
    #         return logsumexp(10000 * a) / 10000

    #     return fn

    def trace_loss(self, dataset: JaxDataset, loss: Loss, smooth=True):
        martingale = convert_jax(self.pre_expr, self.keys_other, smooth=smooth)
        
        if self.use_samples:
            get_fn = build_jax_auxiliary_dataset(dataset, self)
            aux_dataset = get_fn(38)

            #@jit
            def jax_loss(params, i: int):
                batched_eta = vmap(martingale, in_axes=(None, 0))
                eta_pre_matrix_reshaped = batched_eta(params, dataset["pre"]).reshape((-1))
                eta_pre_matrix_reshaped2 = jnp.reshape(eta_pre_matrix_reshaped, (-1, 1))

                batched_batched_eta = vmap(batched_eta, in_axes=(None, 0))
                eta_post_matrix = batched_batched_eta(params, aux_dataset["post"])
                eta_post_matrix_reshaped = eta_post_matrix.reshape(eta_post_matrix.shape[:2])   
                eta_post_matrix_reshaped_weighted = eta_post_matrix_reshaped * aux_dataset["factors"]     
                exp_eta_post_matrix = jnp.mean(eta_post_matrix_reshaped_weighted, axis=1)
              
                max_diff = logsumexp(1000000 * jnp.abs(eta_post_matrix_reshaped - eta_pre_matrix_reshaped2)) / 1000000
                eta_asserting_matrix = batched_eta(params, dataset["asserting"]).reshape((-1))

                penalty_vector_safe = exp_eta_post_matrix - eta_pre_matrix_reshaped + self.epsilon
                penalty_vector_unsafe = -eta_asserting_matrix
                bound_approx = self.get_initial_state_bound_with_limit(params, max_diff, smooth=smooth)

                return loss.compute(penalty_vector_safe, penalty_vector_unsafe, bound_approx)
            return jax_loss
        
        else:
            func = convert_jax(self.post_expr, self.keys_other, smooth=smooth)

            @jit
            def jax_loss(params, i: int):
                batched_eta = vmap(martingale, in_axes=(None, 0))
                eta_pre_matrix_reshaped = batched_eta(params, dataset["pre"]).reshape((-1))
                
                exp_eta_post_matrix = vmap(func, in_axes=(None, 0))(params, dataset["pre"])
                eta_asserting_matrix = batched_eta(params, dataset["asserting"]).reshape((-1))

                penalty_vector_safe = exp_eta_post_matrix - eta_pre_matrix_reshaped + self.epsilon
                penalty_vector_unsafe = -eta_asserting_matrix
                bound_approx = self.get_initial_state_bound_with_limit_expr(params)

                return loss.compute(penalty_vector_safe, penalty_vector_unsafe, bound_approx)

            return jax_loss

def build_jax_auxiliary_dataset(dataset: JaxDataset, problem_data: IsmProblemData) -> JaxAuxiliaryData:
    @jit
    def get_fn(seed):
        random_keys = jrandom.split(jrandom.PRNGKey(seed), SAMPLE_COUNT)
        # random_keys = jrandom.split(jrandom.PRNGKey(seed), SAMPLE_COUNT * dataset["pre"].shape[0])
        # random_keys = jnp.reshape(random_keys, (dataset["pre"].shape[0], SAMPLE_COUNT, 2))
        len_st = len(problem_data.keys)
        len_extra = len(problem_data.keys_special) - len_st
        pre_padded = jnp.pad(dataset["pre"], pad_width=[(0, 0), (0, len_extra)])
        evaluator = JaxEvaluator(problem_data.keys_special, {"p": 0.5})
        get_post_st = evaluator.build_jax_evaluator_root(BlockStatement(statements=problem_data.prog.body))
        def get_post_st_resized(st, key):
            st, key, factor = get_post_st(st, key)
            return jnp.resize(st, (len_st,)), factor
        # post, factors = vmap(vmap(get_post_st_resized, in_axes=(None, 0)), in_axes=(0, 0))(pre_padded, random_keys)
        post, factors = vmap(vmap(get_post_st_resized, in_axes=(None, 0)), in_axes=(0, None))(pre_padded, random_keys)

        return {
            "post": post,
            "factors": factors,
        }
    return get_fn

def build_jax_dataset(dataset: Dataset, problem_data: IsmProblemData) -> JaxDataset:
    return {
        "pre": jnp.asarray(dataset["pre"]),
        "asserting": jnp.asarray(dataset["asserting"]).reshape((-1, len(problem_data.keys)))
    }

def convert_jax(expr, keys, smooth=True):
    @jit
    def expr_jax(params, inputs):
        if isinstance(expr, Number):
            return float(expr)
        elif expr.func == Symbol:
            if expr.name in keys:
                return inputs[keys.index(expr.name)]
            else:
                return params[expr.name]
        elif expr.func == Add:
            result = 0.0
            for summand in expr.args:
                result += convert_jax(summand, keys, smooth)(params, inputs)
            return result
        elif expr.func == Mul:
            result = 1.0
            for multiplicand in expr.args:
                result *= convert_jax(multiplicand, keys, smooth)(params, inputs)
            return result
        elif expr.func == Min and len(expr.args) == 2:
            return jnp.minimum(convert_jax(expr.args[0], keys, smooth)(params, inputs), \
                convert_jax(expr.args[1], keys, smooth)(params, inputs))
        elif expr.func == Max and len(expr.args) == 2:
            return jnp.maximum(convert_jax(expr.args[0], keys, smooth)(params, inputs), \
                convert_jax(expr.args[1], keys, smooth)(params, inputs))
        elif expr.func == soft_function:
            arg = convert_jax(expr.args[0], keys, smooth)(params, inputs)
            if smooth:
                return softminus(arg)
            else:
                return jnp.minimum(arg, SCALE)
        elif expr.func == relu_function:
            arg = convert_jax(expr.args[0], keys, smooth)(params, inputs)
            if smooth:
                return softplus(arg)
            else:
                return relu(arg)
        elif expr.func == Pow:
            arg0 = convert_jax(expr.args[0], keys, smooth)(params, inputs)
            arg1 = convert_jax(expr.args[1], keys, smooth)(params, inputs)
            return jnp.power(arg0, arg1)
        elif expr.func == Piecewise and len(expr.args) == 2 and expr.args[1].cond == True:
            cond = convert_jax(expr.args[0].cond, keys, smooth)(params, inputs)
            arg0 = convert_jax(expr.args[0].expr, keys, smooth)(params, inputs)
            arg1 = convert_jax(expr.args[1].expr, keys, smooth)(params, inputs)
            return jnp.where(cond, arg0, arg1)
        elif expr.func == Lt:
            arg0 = convert_jax(expr.args[0], keys, smooth)(params, inputs)
            arg1 = convert_jax(expr.args[1], keys, smooth)(params, inputs)
            return arg0 < arg1
        elif expr.func == Le:
            arg0 = convert_jax(expr.args[0], keys, smooth)(params, inputs)
            arg1 = convert_jax(expr.args[1], keys, smooth)(params, inputs)
            return arg0 <= arg1
        elif expr.func == Gt:
            arg0 = convert_jax(expr.args[0], keys, smooth)(params, inputs)
            arg1 = convert_jax(expr.args[1], keys, smooth)(params, inputs)
            return arg0 > arg1
        elif expr.func == Ge:
            arg0 = convert_jax(expr.args[0], keys, smooth)(params, inputs)
            arg1 = convert_jax(expr.args[1], keys, smooth)(params, inputs)
            return arg0 >= arg1
        elif expr.func == Eq:
            arg0 = convert_jax(expr.args[0], keys, smooth)(params, inputs)
            arg1 = convert_jax(expr.args[1], keys, smooth)(params, inputs)
            return arg0 >= arg1
        elif expr.func == Abs:
            arg0 = convert_jax(expr.args[0], keys, smooth)(params, inputs)
            return jnp.abs(arg0)
        raise Exception("Conversion Error " + str(expr.func))
    return expr_jax

@dataclass
class MeanBasedLoss(Loss):
    penalty_safe: float
    penalty_unsafe: float
    penalty_initial: float

    def __init__(self, penalty_safe, penalty_unsafe, penalty_initial=1):
        self.penalty_safe = penalty_safe
        self.penalty_unsafe = penalty_unsafe
        self.penalty_initial = penalty_initial

    def compute(self, penalty_safe_vector, penalty_unsafe_vector, initial_state_bound):
        out = (1 / SCALE_FACTOR) * self.penalty_safe * jnp.nan_to_num(jnp.mean((relu(penalty_safe_vector)))) \
            + (1 / SCALE_FACTOR) * self.penalty_unsafe * jnp.nan_to_num(jnp.mean((relu(penalty_unsafe_vector)))) \
                + (1 / SCALE_FACTOR) * self.penalty_initial * initial_state_bound
        return out

    def log(self, config):
        config.loss_type = "mean"
        config.penalty_safe = self.penalty_safe
        config.penalty_unsafe = self.penalty_unsafe

@dataclass
class SumBasedLoss(Loss):
    penalty_safe: float
    penalty_unsafe: float
    penalty_initial: float

    def __init__(self, penalty_safe, penalty_unsafe, penalty_initial=1):
        self.penalty_safe = penalty_safe
        self.penalty_unsafe = penalty_unsafe
        self.penalty_initial = penalty_initial

    def compute(self, penalty_safe_vector, penalty_unsafe_vector, initial_state_bound):
        out = self.penalty_safe * (1 / SCALE_FACTOR) * jnp.sum(relu(penalty_safe_vector)) \
            + self.penalty_unsafe * (1 / SCALE_FACTOR) * jnp.sum(relu(penalty_unsafe_vector)) \
                + self.penalty_initial * initial_state_bound
        return out

    def log(self, config):
        config.loss_type = "sum"
        config.penalty_safe = self.penalty_safe
        config.penalty_unsafe = self.penalty_unsafe

@dataclass
class RefineDatasetStep():
    safe_cxex: int
    unsafe_cxex: int

    def refine(self, params, dataset, problem_data: IsmProblemData, tracker):
        start = time.time()
        pre, asserting = problem_data.get_safe_and_unsafe_cxex(params, self.safe_cxex, self.unsafe_cxex)
        tracker["verify"] += time.time() - start
        print(f"Adding {len(pre)} safe counterexamples")
        print(f"Adding {len(asserting)} unsafe counterexamples")
        print(pre)
        print(asserting)
        return {
            "pre": dataset["pre"] + convert(pre, problem_data.keys_other),
            "asserting": dataset["asserting"] + convert(asserting, problem_data.keys_other)
        }

@dataclass
class Run:
    seed: int
    loss: Loss
    iterations: float
    reuse_params: bool = False
    smooth: bool = True
    learning_rate: float = None # 2e-2
    detailed: bool = False
    save_best: bool = False
    save_best_criteria: float = 0.04
    save_freq: int = 100

    def log(self, dataset: Dataset, config):
        config.iterations = self.iterations
        config.reuse_params = self.reuse_params
        config.learning_rate = self.learning_rate
        config.smooth = self.smooth
        config.safe_states = len(dataset["pre"])
        config.unsafe_states = len(dataset["asserting"])
        self.loss.log(config)

    def get_initial_params(self, params, problem_data: IsmProblemData):
        if self.reuse_params:
            return params
        else:
            key = jrandom.PRNGKey(self.seed)
            if REPSM_INIT:
                initializer = initializers.normal(10)
            else:
                initializer = initializers.normal(0.1)
            rands = initializer(key, (1, len(problem_data.template_vars)))[0]
            initial_params = {}
            for i, param in enumerate(problem_data.template_vars):
                initial_params[param.name] = rands[i]
                if REPSM_INIT and param.name.startswith("b"):
                    initial_params[param.name] += 100
                if param.name.startswith("c"):
                    # initial_params[param.name] += 1
                    initial_params[param.name] *= 10
            return initial_params

    def step(self, loss_func_jax, optimizer):
        def real_step(opt_state, params, i: int):
            l = loss_func_jax(params, i)
            g = grad(loss_func_jax, argnums=0)(params, i)
            updates, opt_state = optimizer.update(g, opt_state)
            params = optax.apply_updates(params, updates)
            return opt_state, params, l
        return real_step

    def train(self, dataset, params, problem_data: IsmProblemData, config, tracker):
        # self.log(dataset, config)
        jax_dataset = build_jax_dataset(dataset, problem_data)
        aux_dataset = build_jax_auxiliary_dataset(jax_dataset, problem_data)(38)
        loss_func_jax = problem_data.trace_loss(jax_dataset, self.loss, smooth=self.smooth)
        if self.learning_rate == None:
            optimizer = optax.rmsprop(4e-4 * SCALE)
        else:
            optimizer = optax.rmsprop(self.learning_rate)
        params = self.get_initial_params(params, problem_data)
        current_opt_state = optimizer.init(params)
        step = jit(self.step(loss_func_jax, optimizer))
        # stat_fn = problem_data.get_stat_info(jax_dataset)

        best_params = None
        best_value = self.save_best_criteria * SCALE

        # if self.save_best:
        #     count = problem_data.get_cxex_count(params, 10, 10)
        #     l = loss_func_jax(params)
        #     if count == (0, 0) and l < best_value:
        #         best_params = params
        #         best_value = l

        print("Safe States:", str(len(dataset["pre"])) + ",", "Unsafe States:", len(dataset["asserting"]))
        print("")
        for i in range(self.iterations):
            old_params = params
            start = time.time()
            current_opt_state, params, l = step(current_opt_state, params, i)
            tracker["learn"] += time.time() - start
            if i % 100 == 0:
                # print(stat_fn(old_params))
                if self.detailed:
                    start = time.time()
                    count = problem_data.get_cxex_count(old_params, 10, 10)
                    tracker["verify"] += time.time() - start
                    info = problem_data.get_helper_info_other(old_params, jax_dataset, aux_dataset)
                    # print(info)
                    st_bound = problem_data.get_initial_state_bound(old_params, info)
                    if self.save_best and count == (0, 0) and st_bound < best_value:
                        best_params = old_params
                        best_value = st_bound
                    print("iteration =", i, "loss =", l, "cxex =", count, "st_bound =", st_bound)
                else:
                    print("iteration =", i, "loss =", l)
                #  #, "cxex = ", len(get_counterexamples_small(params)))
            # else:
            #     print("iteration =", i, "loss =", l)
        if self.save_best:
            count = problem_data.get_cxex_count(params, 10, 10)
            info = problem_data.get_helper_info_other(params, jax_dataset, aux_dataset)
            st_bound = problem_data.get_initial_state_bound(params, info)
            if count == (0, 0) and st_bound < best_value:
                best_params = params
                best_value = st_bound
            if best_params != None:
                params = best_params
                l = loss_func_jax(params, i)
        print("final loss =", l) #, "cxex = ", len(get_counterexamples_small(params)))
        count = problem_data.get_cxex_count(params, 50, 50)
        print("final cxex =", count)
        print()
        print(show(params))
        print()
        pprint(show(self.get_stats(jax_dataset, problem_data, params)))
        print()
        # (safe_cxex_dataset, unsafe_cxex_dataset) = self.get_stats(jax_dataset, problem_data, params)
        # print("dataset cxex =", (int(safe_cxex_dataset), int(unsafe_cxex_dataset)))
        return params

    # Warning: not ready to be used
    def get_stats(self, dataset: JaxDataset, problem_data: IsmProblemData, params):
        get_fn = build_jax_auxiliary_dataset(dataset, problem_data)
        aux_dataset = get_fn(38)
        func = convert_jax(problem_data.post_expr, problem_data.keys_other, smooth=False)
        martingale = convert_jax(problem_data.pre_expr, problem_data.keys_other, smooth=False)
        
        batched_eta = vmap(martingale, in_axes=(None, 0))
        eta_pre_matrix_reshaped = batched_eta(params, dataset["pre"]).reshape((-1))
        exp_eta_post_matrix = vmap(func, in_axes=(None, 0))(params, dataset["pre"])
        eta_asserting_matrix = batched_eta(params, dataset["asserting"]).reshape((-1))
        penalty_safe_vector = exp_eta_post_matrix - eta_pre_matrix_reshaped
        # penalty_unsafe_vector = SCALE - eta_asserting_matrix
        penalty_unsafe_vector = -eta_asserting_matrix

        # print(dataset["pre"][penalty_safe_vector > 0])

        batched_batched_eta = vmap(batched_eta, in_axes=(None, 0))
        eta_post_matrix = batched_batched_eta(params, aux_dataset["post"])
        eta_post_matrix_reshaped = eta_post_matrix.reshape(eta_post_matrix.shape[:2])   
        eta_post_matrix_reshaped_weighted = eta_post_matrix_reshaped * aux_dataset["factors"]     
        other_exp_eta_post_matrix = jnp.mean(eta_post_matrix_reshaped_weighted, axis=1)
        other_penalty_safe_vector = other_exp_eta_post_matrix - eta_pre_matrix_reshaped

        return {
            "count_penalty_safe": jnp.sum(penalty_safe_vector > 0),
            "count_penalty_unsafe": jnp.sum(penalty_unsafe_vector > 0),
            "count_penalty_safe_proxy": jnp.sum(other_penalty_safe_vector > 0),
            "count_same": jnp.sum((penalty_safe_vector > 0) & (other_penalty_safe_vector > 0)),
            "count_penalty_safe_proxy1": jnp.sum(other_penalty_safe_vector > 0.1),
            "sum_penalty_unsafe": jnp.sum(relu(penalty_unsafe_vector)),
            "sum_penalty_safe": jnp.sum(relu(penalty_safe_vector)),
            "sum_penalty_safe_proxy": jnp.sum(relu(other_penalty_safe_vector)),
            "mean_penalty_unsafe": jnp.mean(relu(penalty_unsafe_vector)),
            "mean_penalty_safe": jnp.mean(relu(penalty_safe_vector)),
            "mean_penalty_safe_proxy": jnp.mean(relu(other_penalty_safe_vector)),
        }

def convert(dataset, keys):
    result = []
    for item in dataset:
        obj = []
        for key in keys:
            obj.append(item[key])
        result.append(obj)
    return result

def default_is_asserting(st):
    return st["a"] == 1

def build_real_dataset(prog: WhileLoopProgram, problem_data: IsmProblemData, args, is_asserting=default_is_asserting, count=5) -> Dataset:
    evaluator = Evaluator()
    pre = []
    for _ in range(0, count):
        pre += evaluator.evaluate_while_loop_program(prog, args, 100, is_asserting=is_asserting)
    return {
        "pre": convert(pre, problem_data.keys),
        "asserting": []
    }

def build_synthetic_dataset() -> Dataset:
    keys = ["red", "blue", "a"]

    def is_terminal(st):
        return st["red"] <= 0 and st["blue"] <= 0

    result = {"pre": [], "asserting": []}
    for i in range(-5, 6):
        for j in range(-5, 6):
            st = {"red": i, "blue": j, "a": 0}
            st_asserting = {"red": i, "blue": j, "a": 1}
            if not is_terminal(st):
                result["pre"].append(st)
                #result["asserting"].append(st_asserting)
    result["pre"] = convert(result["pre"], keys)
    result["asserting"] = convert(result["asserting"], keys)
    return result

@dataclass
class LearningTask:
    runs: List[Run]
    def get_separator(self):
        return "========================================"

    def go(self, path: str, dataset: Dataset, problem_data: IsmProblemData):
        # name = self.get_name(path)
        params = None
        tracker = {"learn": 0, "verify": 0}
        # print(f"Task Name: {name}")s
        for i, step in enumerate(self.runs):
            print()
            print(self.get_separator())
            print(f"Beginning Step {i + 1}/{len(self.runs)}")
            print()
            if isinstance(step, Run):
                params = step.train(dataset, params, problem_data, None, tracker)
            elif isinstance(step, RefineDatasetStep):
                dataset = step.refine(params, dataset, problem_data, tracker)
        return params, tracker

@dataclass
class LearningExperiment:
    task: LearningTask
    tests: int

    def get_suffix(self):
        return "".join(random.choice(string.ascii_uppercase + string.digits) for _ in range(5))

    def get_name(self, path):
        return f"{Path(path).stem}_{self.get_suffix()}"

    def go(self, path: str, dataset: Dataset, problem_data: IsmProblemData):
        dir_name = datetime.today().strftime("%Y-%m-%d")
        if "RESULTS_SUFFIX" in os.environ:
            suffix = os.environ["RESULTS_SUFFIX"]
            results = Path(f"results-{suffix}")
        else:
            results = Path("results")
        if not results.is_dir():
            print("Error: Results directory not found")
            return
        name = datetime.today().strftime(f"%H_%M_%S_{Path(path).stem}")
        (results / dir_name).mkdir(exist_ok=True)
        workspace = results / dir_name / name
        workspace.mkdir()
        print(f"Task Name: {name}")
        martingale = convert_jax(problem_data.pre_expr, problem_data.keys_other, smooth=False)
        bounds = []
        learn_times = []
        verify_times = []
        with (workspace / "overview.txt").open("w") as overview:
            with (workspace / "config.txt").open("w") as file:
                pprint(self.task, stream=file)
                print("Configuration File Saved")
            for i in range(self.tests):
                print(f"Running Test {i}")
                with (workspace / f"test{i}.txt").open("w", buffering=1) as file:
                    with redirect_stdout(file):
                        params, tracker = self.task.go(path, dataset, problem_data)
                info = problem_data.get_helper_info(params, dataset)
                if not problem_data.is_really_valid(params, info):
                    overview.write(f"Test {i}, Invalid\n")
                else:
                    bound = problem_data.get_real_bound(params, info)
                    # bound = float(martingale(params, problem_data.initial_state)) / SCALE
                    bounds.append(bound)
                    learn_time = tracker["learn"]
                    verify_time = tracker["verify"]
                    learn_times.append(learn_time)
                    verify_times.append(verify_time)
                    overview.write(f"Test {i}, Bound: {bound}, Learn Time: {learn_time}, Verify Time: {verify_time}\n")
            if len(bounds) != 0:
                overview.write(f"Average Bound: {statistics.mean(bounds)}\n")
                overview.write(f"Average Learn Time: {statistics.mean(learn_times)}\n")
                overview.write(f"Average Verify Time: {statistics.mean(verify_times)}\n")

class Configuration(ABC):
    @abstractmethod
    def get_path(self):
        pass

    @abstractmethod
    def get_task(self):
        pass

    @abstractmethod
    def get_problem_data(self, prog: WhileLoopProgram):
        pass

    @abstractmethod
    def get_dataset(self):
        pass

    def get_all(self):
        path = self.get_path()
        task = self.get_task()
        prog = load_program(path)
        prog = transform_program(prog)
        problem_data = self.get_problem_data(prog)
        dataset = self.get_dataset(prog, problem_data)
        return path, dataset, problem_data, task

def fix(expr):
    arg = Symbol("arg")
    return expr.replace(relu_function, Lambda(arg, Max(arg, 0))) \
        .replace(soft_function, Lambda(arg, Min(arg, SCALE)))
