from ast import Expression
from enum import Enum
from typing import Dict, List, Tuple, Union, TypedDict
import operator
import numpy as np
from scipy import stats
import random
from sympy import Idx
import z3
import jax.numpy as jnp
import jax.lax as lax
import jax.random as jrandom
from fractions import Fraction
from lib.typing import *

BINARY_OPERATORS = {
    "+": operator.add,
    "-": operator.sub,
    "*": operator.mul,
    "==": operator.eq,
    "<": operator.lt,
    "<=": operator.le,
    ">": operator.gt,
    ">=": operator.ge,
    "&&": operator.and_,
    "||": operator.or_
}

SamplingRemapper = Dict[str, float]

class JaxEvaluator:
    """
        Statements: (st, key, factor) -> (st, key, factor)
        Expressions: st -> value

        CS = CombinedState = (st, key, factor)

        N.B. SamplingRemapper only for Bernoulli distributions
    """

    def __init__(self, keys: List[str], remapper: SamplingRemapper = {}):
        self.keys = keys
        self.remapper = remapper

    def build_jax_evaluator_expr(self, expr):
        match expr:
            case IntNode(value=value):
                return lambda st: value
            case RealNode(value=value):
                return lambda st: float(value)
            case BoolNode(value=value):
                return lambda st: value
            case IdentifierNode(name=name):
                idx = self.keys.index(name)
                return lambda st: st[idx]
            case BinaryOperator(operator=operator, operand1=operand1, operand2=operand2):
                fn1 = self.build_jax_evaluator_expr(operand1) 
                fn2 = self.build_jax_evaluator_expr(operand2)
                return lambda st: BINARY_OPERATORS[operator](fn1(st), fn2(st))
        raise Exception("Build Jax Evaluator Error")

    def build_jax_evaluator_stmt(self, stmt):
        match stmt:
            case AssignmentStatement(
                name=name, 
                expr=UnaryOperator(operator="sample", operand=DistributionNode(operator=operator, args=args))
            ):
                idx = self.keys.index(name)
                args_fn = [self.build_jax_evaluator_expr(arg) for arg in args]
                # print("assignment sample")
                match (operator, args_fn):
                    case ("Bernoulli", [p_fn]):
                        if name in self.remapper:
                            fake_p = self.remapper[name]
                            def result(st, key, factor):
                                next_key, key = jrandom.split(key)
                                p = p_fn(st)
                                bern = jrandom.bernoulli(key, p=fake_p)
                                next_factor = jnp.where(bern == 1, p / fake_p, (1 - p) / (1 - fake_p)) * factor
                                return st.at[idx].set(bern), next_key, next_factor
                        else:
                            def result(st, key, factor):
                                next_key, key = jrandom.split(key)
                                return st.at[idx].set(jrandom.bernoulli(key, p=p_fn(st))), next_key, factor
                        return result
                    case ("Uniform", [p_fn, q_fn]):
                        def result(st, key, factor):
                            next_key, key = jrandom.split(key)
                            return st.at[idx].set(jrandom.uniform(key, minval=p_fn(st), maxval=q_fn(st))), next_key, factor
                        return result
                    case ("Gaussian", [p_fn, q_fn]):
                        def result(st, key, factor):
                            next_key, key = jrandom.split(key)
                            return st.at[idx].set(p_fn(st) + jrandom.normal(key) * q_fn(st)), next_key, factor
                        return result
                    case _:
                        raise Exception("Not sure what to do")

            case AssignmentStatement(name=name, expr=expr):
                # print("assignment regular")
                fn = self.build_jax_evaluator_expr(expr)
                idx = self.keys.index(name)
                return lambda st, key, factor: (st.at[idx].set(fn(st)), key, factor)
            case IfStatement(branches=branches):
                # print("if statement")
                if len(branches) == 2 and branches[1][0] == BoolNode(True):
                    (cond1, block1) = branches[0]
                    (cond2, block2) = branches[1]
                    cond1_fn = self.build_jax_evaluator_expr(cond1)
                    # cond2_fn = build_jax_evaluator_expr(keys, cond2)
                    block1_fn = self.build_jax_evaluator(block1)
                    block2_fn = self.build_jax_evaluator(block2)
                    def func(st, key, factor):
                        idx = cond1_fn(st)
                        return lax.cond(idx, block1_fn, block2_fn, st, key, factor)
                    return func
                elif len(branches) == 1:
                    (cond1, block1) = branches[0]
                    cond1_fn = self.build_jax_evaluator_expr(cond1)
                    block1_fn = self.build_jax_evaluator(block1)
                    def id_fn(st, key, factor):
                        return st, key, factor
                    def func(st, key, factor):
                        idx = cond1_fn(st)
                        return lax.cond(idx, block1_fn, id_fn, st, key, factor)
                    return func
                else:
                    raise Exception("Not sure what to do")
                # branches_reversed = list(branches)
                # branches_reversed.reverse()
                # else_fn = lambda st, key: (st, key)
                # for (cond, block) in branches_reversed:
                #     cond_fn = build_jax_evaluator_expr(keys, cond)
                #     block_fn = build_jax_evaluator(keys, block)
                #     print("result")
                #     save_else = else_fn
                #     def next_result(st, key):
                #         print("next result", cond, st, key)
                #         idx = cond_fn(st)
                #         return lax.cond(idx, block_fn, lambda st, key: (st, key), st, key)
                #     else_fn = next_result
                # return else_fn
            case _:
                raise Exception("Not sure what to do")

    def build_jax_evaluator_root(self, block: BlockStatement):
        fn = self.build_jax_evaluator(block)
        return lambda st, key: fn(st, key, 1)

    def build_jax_evaluator(self, block: BlockStatement):
        fns = list(map(lambda stmt: self.build_jax_evaluator_stmt(stmt), block.statements))
        
        def evaluate(*cs):
            for fn in fns:
                # print("here in")
                cs = fn(*cs)
                # print("here out")
            return cs
        return evaluate

class Evaluator:
    def __init__(self, seed=42):
        self.rng = np.random.default_rng(seed)

    def evaluate_expr(self, expr: Expression, env):
        match expr:
            case IntNode(value=value):
                return value
            case RealNode(value=value):
                return float(value)
            case BoolNode(value=value):
                return value
            case IdentifierNode(name=name):
                return env[name]
            case UnaryOperator(operator="sample", operand=DistributionNode(operator=operator, args=args)):
                args = [self.evaluate_expr(arg, env) for arg in args]
                match (operator, args):
                    case ("Bernoulli", [p]):
                        return int(stats.bernoulli(float(p)).rvs(size=1, random_state=self.rng))
                    case ("Uniform", [p,q]):
                        return float(stats.uniform(p, q - p).rvs(size=1, random_state=self.rng))
                    case ("Gaussian", [p,q]):
                        return float(stats.norm(p, q).rvs(size=1, random_state=self.rng))
            case BinaryOperator(operator=operator, operand1=operand1, operand2=operand2):
                val1 = self.evaluate_expr(operand1, env)
                val2 = self.evaluate_expr(operand2, env)
                return BINARY_OPERATORS[operator](val1, val2)

    def evaluate_statement(self, stmt: Statement, env):
        match stmt:
            case AssignmentStatement(name=name, expr=expr):
                val = self.evaluate_expr(expr, env)
                env[name] = val
                #print(env)
                #print(f"{name} = {val}")
            case IfStatement(branches=branches):
                env_keys = set(env)
                for (cond, block) in branches:
                    if self.evaluate_expr(cond, env):
                        self.evaluate_block(block, env)
                        break
                unwanted_keys = set(env) - env_keys
                for unwanted_key in unwanted_keys:
                    del env[unwanted_key]
            case WhileStatement(guard=guard, body=body):
                env_keys = set(env)
                while self.evaluate_expr(guard, env):
                    self.evaluate_block(body, env)
                unwanted_keys = set(env) - env_keys
                for unwanted_key in unwanted_keys:
                    del env[unwanted_key]

    def evaluate_block(self, block: BlockStatement, env):
        for stmt in block.statements:
            self.evaluate_statement(stmt, env)

    def evaluate_function(self, func: Function, args):
        env = self.init_env(func.args, args)
        self.evaluate_block(func.body, env)

    def error(self):
        raise Exception("Evaluator Error")

    def check_args(self, arg_spec, args):
        if len(arg_spec) != len(args):
            self.error()
        for ((_, arg_type), arg) in zip(arg_spec, args):
            if arg_type == PrimType.INT and isinstance(arg, int):
                continue
            if arg_type == PrimType.REAL and isinstance(arg, float):
                continue
            self.error()

    def init_env(self, arg_spec, args):
        env = {}
        for ((arg_name, _), arg) in zip(arg_spec, args):
            env[arg_name] = arg
        return env

    def evaluate(self, prog: Program, args):
        for function in prog.functions:
            if function.name == "main":
                self.check_args(function.args, args)
                self.evaluate_function(function, args)
                break
    
    def evaluate_while_loop_program(self, prog: WhileLoopProgram, args, max_iters=None, include_terminal=False, is_asserting=None):
        self.check_args(prog.args, args)
        history = []
        env = self.init_env(prog.args, args)
        init_block = BlockStatement(prog.init)
        update_block = BlockStatement(prog.body)
        self.evaluate_block(init_block, env)
        i = 0
        env_keys = set(env)
        while max_iters == None or i < max_iters:
            if not self.evaluate_expr(prog.guard, env):
                if include_terminal:
                    history.append(env.copy())
                break
            if is_asserting != None and is_asserting(env):
                break
            history.append(env.copy())
            self.evaluate_block(update_block, env)
            unwanted_keys = set(env) - env_keys
            for unwanted_key in unwanted_keys:
                del env[unwanted_key]
            i += 1
        
        return history

    def get_successor(self, prog: WhileLoopProgram, env, branching_factor):
        update_block = BlockStatement(prog.body)
        env_keys = set(env)
        next = []
        for j in range(branching_factor):
            env_copy = env.copy()
            self.evaluate_block(update_block, env_copy)
            unwanted_keys = set(env_copy) - env_keys
            for unwanted_key in unwanted_keys:
                del env_copy[unwanted_key]
            next.append(env_copy)
        return next

    def evaluate_while_loop_program_branching(self, prog: WhileLoopProgram, args, max_iters=None, branching_factor=200):
        self.check_args(prog.args, args)
        pre = []
        post = []
        env = self.init_env(prog.args, args)
        init_block = BlockStatement(prog.init)
        update_block = BlockStatement(prog.body)
        self.evaluate_block(init_block, env)
        i = 0
        env_keys = set(env)
        while max_iters == None or i < max_iters:
            if not self.evaluate_expr(prog.guard, env):
                break
            
            next = []
            for j in range(branching_factor):
                env_copy = env.copy()
                self.evaluate_block(update_block, env_copy)
                unwanted_keys = set(env_copy) - env_keys
                for unwanted_key in unwanted_keys:
                    del env_copy[unwanted_key]
                next.append(env_copy)

            pre.append(env.copy())
            post.append(next)
            
            env = next[0]
            i += 1
        return pre, post
