from dataclasses import replace
from lib.parser import TypeChecker
from typing import Dict, List, Tuple, Union
from sympy import Add, And, DiracDelta, Eq, Float, Integer, Lt, Max, Mul, Or, Piecewise, Rational, Symbol, exp, false, integrate, oo, pi, sqrt, true, solve, Ge, Le, Gt, simplify, Min, Pow, Ne, Integral, Lambda, Abs, expand
from sympy.logic.boolalg import BooleanTrue, BooleanFalse
import sympy
from lib.typing import *
import z3

def is_exact(expr, var):
    return expr == var or expr == -var

def contains_exact(expr, var):
    if is_exact(expr, var):
        return True
    elif expr.func != Add:
        return False
    args = [arg for arg in expr.args if arg.has(var)]
    return len(args) == 1 and is_exact(args[0], var)

def general_integrate(expr, var):
    if not expr.has(DiracDelta):
        return integrate(expr, (var, -oo, oo))
    else:
        return delta_integrate(expr, var)

def delta_integrate(expr, var):
    if expr.func == Mul:
        for factor in expr.args:
            if factor.func == DiracDelta and contains_exact(factor.args[0], var):
                replacement = solve(factor.args[0], var)[0]
                # print(replacement)
                other_factors = [other_factor.subs(var, replacement) for other_factor in expr.args if other_factor != factor]
                return Mul(*other_factors)
        for factor in expr.args:
            if factor.func == Add:
                check = True
                for summand in factor.args:
                    if not summand.has(var):
                        check = False
                        break
                if not check:
                    continue
                other_factors = [other_factor for other_factor in expr.args if other_factor != factor]
                other_factors = Mul(*other_factors)
                expanded_factors = [summand * other_factors for summand in factor.args]
                expanded_integrated = [delta_integrate(factor, var) for factor in expanded_factors]
                if None not in expanded_integrated:
                    return Add(*expanded_integrated)
            elif factor.func == Piecewise:
                check = True
                for expr_cond_pair in factor.args:
                    if not expr_cond_pair.expr.has(var):
                        check = False
                        break
                if not check:
                    continue
                other_factors = [other_factor for other_factor in expr.args if other_factor != factor]
                other_factors = Mul(*other_factors)
                conds = [pair.cond for pair in factor.args]
                expanded_factors = [pair.expr * other_factors for pair in factor.args]
                expanded_integrated = [delta_integrate(factor, var) for factor in expanded_factors]
                if None not in expanded_integrated:
                    return Piecewise(*zip(expanded_integrated, conds))
    elif expr.func == Add:
        summands_integrated = [delta_integrate(summand, var) for summand in expr.args]
        if None not in summands_integrated:
            return Add(*summands_integrated)
    elif expr.func == Piecewise:
        conds = [pair.cond for pair in expr.args]
        expanded_integrated = [delta_integrate(pair.expr, var) for pair in expr.args]
        if None not in expanded_integrated:
            return Piecewise(*zip(expanded_integrated, conds))
    return None

def compute_expectation(expr, var):
    return integrate(expr * var, (var, -oo, oo))

class VariableClass(Enum):
    PRIOR = 0
    DISCRETE = 1
    CONTINUOUS = 2

def translate(var, number):
    return Symbol(f"{var}_{number}")

@dataclass
class Dist:
    pdf: sympy.Expr
    vars_map: Dict[str, int]
    vars_type: Dict[Tuple[str, int], VariableClass]
    begin_map: Dict[str, int]
    vars_list: List[Tuple[str, int]]
    vars: Set[str]
    begin_type_map: Dict[str, PrimType]

    def delta_integrate(self, var):
        return replace(self, pdf=delta_integrate(self.pdf, var))

    def add(self, var, expr):
        return replace(self, pdf=self.pdf * DiracDelta(var - expr))

    def multiply(self, expr):
        return replace(self, pdf=self.pdf * expr)

    def get_var(self, var):
        return translate(var, self.vars_map[var])

    def add_var_base(self, var, var_class, next, t=None):
        vars_map = self.vars_map.copy()
        vars_type = self.vars_type.copy()
        vars_list = self.vars_list.copy()
        begin_type_map = self.begin_type_map.copy()

        vars_map[var] = next
        vars_type[(var, next)] = var_class
        vars_list.append((var, next))
        if t != None:
            begin_type_map[var] = t

        result = replace(self, vars_map=vars_map, vars_type=vars_type, vars_list=vars_list, begin_type_map=begin_type_map)
        return result, result.get_var(var)

    def add_var(self, var, var_class, t=None):
        next = 0
        if var in self.vars_map:
            next = self.vars_map[var] + 1
        return self.add_var_base(var, var_class, next, t=t)

    @staticmethod
    def create(vars_map, begin_type_map):
        vars_map = vars_map.copy()
        vars_type = {(k, v): VariableClass.PRIOR for k, v in vars_map.items()}
        begin_map = vars_map.copy()
        return Dist(pdf=Integer(1), vars_map=vars_map, vars_type=vars_type, begin_map=begin_map, vars_list=[], vars=set(), begin_type_map=begin_type_map)

    def integrate(self, var):
        symbol = translate(var[0], var[1])
        vars_type = self.vars_type.copy()
        vars_list = self.vars_list.copy()
        del vars_type[var]
        vars_list.remove(var)
        # if self.pdf.has(DiracDelta):
        return replace(self, pdf=general_integrate(self.pdf, symbol), vars_list=vars_list, vars_type=vars_type)
        # else:
        #     return replace(self, pdf=integrate(self.pdf, (symbol, -oo, oo)), vars_list=vars_list, vars_type=vars_type)

    def integrate_many(self, vars):
        dist = self
        for var in vars:
            dist = dist.integrate(var)
        return dist

    def integrate_many_smart(self, vars):
        vars_reordered = []
        for var in vars:
            if self.vars_type[var] == VariableClass.DISCRETE:
                vars_reordered.append(var)
        for var in vars:
            if self.vars_type[var] != VariableClass.DISCRETE:
                vars_reordered.append(var)
        return self.integrate_many(vars_reordered)

    def produce_child(self):
        return Dist.create(self.vars_map, self.begin_type_map)

    def _repr_latex_(self):
        return self.pdf._repr_latex_()
    
    def simplify(self):
        return replace(self, pdf=simplify(self.pdf))

    def get_post(self, expr):
        dist = self.add_result(expr)
        dist = dist.integrate_many_smart(dist.vars_list)
        return dist.pdf

    def add_result(self, expr):
        dist, result = self.add_var("result", VariableClass.DISCRETE, t=PrimType.REAL)
        for var in dist.begin_map:
            expr = expr.subs(Symbol(f"{var}_{dist.begin_map[var]}"), 
                Symbol(f"{var}_{dist.vars_map[var]}"))
        dist = dist.add(result, expr) \
            .multiply(result)
        return dist
    
    def add_result_partial(self, expr):
        dist, result = self.add_var("result", VariableClass.DISCRETE, t=PrimType.REAL)
        for var in dist.begin_map:
            expr = expr.subs(Symbol(f"{var}_{dist.begin_map[var]}"), 
                Symbol(f"{var}_{dist.vars_map[var]}"))
        dist = dist.add(result, expr)
        return dist

class Analyser:
    def analyse(self, prog: WhileLoopProgram):
        checker = TypeChecker()
        env = checker.check_args(prog.args)
        env = checker.check_block(env, BlockStatement(prog.init))
        # Why not use env, to create vars?
        vars = self.get_assigned(prog.init)
        vars.update([name for (name, _) in prog.args])
        # print(vars)
        vars_map = {var: 0 for var in vars}
        # begin_map = vars_map.copy()
        dist = Dist.create(vars_map, env)
        return self.translate_expr(dist, prog.guard), self.produce(dist, prog.body)

    def translate_expr_real(self, dist: Dist, expr: Expr):
        match expr:
            case BoolNode(value=value):
                return (true if value else false), PrimType.BOOL
            case IntNode(value=value):
                return Integer(value), PrimType.INT
            case RealNode(value=value):
                return Rational(value), PrimType.REAL
            case IdentifierNode(name=name):
                return dist.get_var(name), dist.begin_type_map[name]
            case BinaryOperator(operator=operator, operand1=operand1, operand2=operand2):
                mapping = {
                    "==": Eq,
                    "+": Add,
                    "-": lambda x, y: x - y,
                    "*": Mul,
                    "<": Lt,
                    ">": Gt,
                    "||": Or,
                    "&&": And,
                    ">=": Ge,
                    "<=": Le
                }
                if operator in mapping:
                    op1, t1 = self.translate_expr_real(dist, operand1)
                    op2, t2 = self.translate_expr_real(dist, operand2)
                    return mapping[operator](op1, op2), t1
                else:
                    print(operator)
                    raise Exception("Not sure what to do")
            case _:
                raise Exception("Not sure what to do")

    def translate_expr(self, dist: Dist, expr: Expr):
        return self.translate_expr_real(dist, expr)[0]

    def produce(self, dist: Dist, statements: List[Statement]):
        # print(pdf)
        for stmt in statements:
            match stmt:
                case AssignmentStatement(
                    name=name, 
                    expr=UnaryOperator(operator="sample", operand=DistributionNode(operator=operator, args=args))
                ):
                    args_translated = [self.translate_expr(dist, arg) for arg in args]
                    match (operator, args_translated):
                        case ("Bernoulli", [p]):
                            dist, name_translated = dist.add_var(name, VariableClass.DISCRETE, t=PrimType.INT)
                            sub_pdf = p * DiracDelta(name_translated - 1) + (1 - p) * DiracDelta(name_translated)
                            dist = dist.multiply(sub_pdf)
                        case ("Uniform", [p, q]):
                            dist, name_translated = dist.add_var(name, VariableClass.CONTINUOUS, t=PrimType.REAL)
                            sub_pdf = Piecewise((Integer(1) / (q - p), And(Ge(name_translated, p), Le(name_translated, q))), (0, True))
                            dist = dist.multiply(sub_pdf)
                        case ("Gaussian", [p, q]):
                            dist, name_translated = dist.add_var(name, VariableClass.CONTINUOUS, t=PrimType.REAL)
                            sub_pdf = (Integer(1) / (q * sqrt(2 * pi))) * exp(- (Integer(1)/Integer(2)) * ((name_translated - p) / q) ** 2)
                            dist = dist.multiply(sub_pdf)
                        case _:
                            raise Exception("Not sure what to do")
                case AssignmentStatement(
                    name=name,
                    expr=expr
                ):
                    expr_translated, t = self.translate_expr_real(dist, expr)
                    dist, name_translated = dist.add_var(name, VariableClass.DISCRETE, t=t)
                    dist = dist.add(name_translated, expr_translated)
                case IfStatement(branches=branches_original):
                    branches = branches_original + [(BoolNode(True), BlockStatement(statements=[]))]
                    piecewise = []

                    for (cond, body) in branches:
                        piecewise.append((self.produce(dist.produce_child(), body.statements), self.translate_expr(dist, cond)))
                    
                    max_map = {}
                    for key, value in dist.vars_map.items():
                        count = value
                        for (sub_dist, _) in piecewise:
                            count = max(sub_dist.vars_map[key], count)
                        if value != count:
                            max_map[key] = count + 1

                    for key, value in max_map.items():
                        dist, _ = dist.add_var_base(key, VariableClass.DISCRETE, value)

                    piecewise2 = []
                    for (sub_dist, cond) in piecewise:
                        for key in max_map:
                            sub_dist = sub_dist.add(dist.get_var(key), sub_dist.get_var(key))
                        # for var, number in sub_dist.vars_list:
                        #     sub_dist = sub_dist.delta_integrate(translate(var, number))
                        sub_dist = sub_dist.integrate_many(sub_dist.vars_list)
                        piecewise2.append((sub_dist.pdf, cond))

                    dist = dist.multiply(Piecewise(*piecewise2))
                case _:
                    raise Exception("Not sure what to do")
            # print(pdf)
        return dist

    def error(self):
        raise Exception("Analyser Error")

    def get_assigned(self, statements: List[Statement]):
        vars = set()
        for stmt in statements:
            match stmt:
                case AssignmentStatement(name=name, expr=_):
                    vars.add(name)
        return vars

def z3_max(x, y):
    return z3.If(x > y, x, y)

def z3_min(x, y):
    return z3.If(x < y, x, y)

def z3_abs(x):
    return z3.If(x > 0, x, -x)

def convert_expr(expr, dist: Dist, vars: Set[str]):
    if isinstance(expr, sympy.Integer):
        return z3.Q(int(expr), 1)
    if isinstance(expr, Rational):
        return z3.Q(expr.numerator, expr.denominator)
    elif expr.func == Symbol:
        if expr.name in vars:
            return z3.Real(expr.name)
        prefix = expr.name.split("_")[0]
        var_type = dist.begin_type_map[prefix]
        if var_type == PrimType.INT:
            return z3.Int(expr.name)
        elif var_type == PrimType.REAL:
            return z3.Real(expr.name)
        else:
            print(expr.name)
            raise Exception("Conversion Error")
        # if expr.is_Integer:
        #    return z3.Int(expr.name)
        #else:
        #    return z3.Real(expr.name)
    elif expr.func == Add:
        return z3.Sum([convert_expr(arg, dist, vars) for arg in expr.args])
    elif expr.func == Mul:
        return z3.Product([convert_expr(arg, dist, vars) for arg in expr.args])
    elif expr.func == Piecewise and len(expr.args) >= 2:
        size = len(expr.args)
        if expr.args[size - 1].cond.func != BooleanTrue:
            raise Exception("Conversion Error")
        args = [convert_expr(pair.expr, dist, vars) for pair in expr.args]
        conds = [convert_equation(pair.cond, dist, vars) for pair in expr.args]
        curr = args[size - 1]
        for j in range(size - 2, -1, -1):
            curr = z3.If(conds[j], args[j], curr)
        return curr
    elif expr.func == Max and len(expr.args) == 2:
        arg1 = convert_expr(expr.args[0], dist, vars)
        arg2 = convert_expr(expr.args[1], dist, vars)
        return z3_max(arg1, arg2)
    elif expr.func == Min and len(expr.args) == 2:
        arg1 = convert_expr(expr.args[0], dist, vars)
        arg2 = convert_expr(expr.args[1], dist, vars)
        return z3_min(arg1, arg2)
    elif expr.func == Pow and expr.args[1] == 2:
        arg1 = convert_expr(expr.args[0], dist, vars)
        return arg1 * arg1
    elif expr.func == Pow and expr.args[1] == -1:
        arg1 = convert_expr(expr.args[0], dist, vars)
        return 1 / arg1
    elif expr.func == Abs:
        arg1 = convert_expr(expr.args[0], dist, vars)
        return z3_abs(arg1)
    print(expr.func)
    raise Exception("Conversion Error")

def convert_equation(eq, dist: Dist, vars: Set[str]):
    if eq.func == Ge:
        return convert_expr(eq.lhs, dist, vars) >= convert_expr(eq.rhs, dist, vars)
    elif eq.func == Le:
        return convert_expr(eq.lhs, dist, vars) <= convert_expr(eq.rhs, dist, vars)
    elif eq.func == Gt:
        return convert_expr(eq.lhs, dist, vars) > convert_expr(eq.rhs, dist, vars)
    elif eq.func == Lt:
        return convert_expr(eq.lhs, dist, vars) < convert_expr(eq.rhs, dist, vars)
    elif eq.func == Eq:
        return convert_expr(eq.lhs, dist, vars) == convert_expr(eq.rhs, dist, vars)
    elif eq.func == And:
        return z3.And(*{convert_equation(arg, dist, vars) for arg in eq.args})
    elif eq.func == BooleanTrue:
        return True
    elif eq.func == BooleanFalse:
        return False
    elif eq.func == Or:
        return z3.Or(*{convert_equation(arg, dist, vars) for arg in eq.args})
    print(eq.func)
    raise Exception("Conversion Error")

def convert_from_prog_expr(expr, map):
    match expr:
        case IntNode(value=value):
            return value
        case RealNode(value=value):
            return z3.Q(value.numerator, value.denominator)
        case IdentifierNode(name=name):
            return z3.Int(f"{name}_{map[name]}")
        case BinaryOperator(operator=operator, operand1=operand1, operand2=operand2):
            expr1 = convert_from_prog_expr(operand1, map)
            expr2 = convert_from_prog_expr(operand2, map)
            if operator == "||":
                return z3.Or(expr1, expr2)
            elif operator == ">":
                return expr1 > expr2
            elif operator == "<":
                return expr1 < expr2
            elif operator == "<=":
                return expr1 <= expr2
            elif operator == ">=":
                return expr1 >= expr2
    raise Exception("Conversion Error")

class Indicator(sympy.Expr):
    def _numpycode(self, printer):
        return printer._print(self.args[0])

def get_indicator(expr):
    if expr.func == Piecewise:
        if expr.args[1].expr != 0 or expr.args[1].cond != True:
            return None
        return expr.args[0].expr * Indicator(expr.args[0].cond)
    elif expr.func == Max:
        if expr.args[0] != 0:
            return None
        return Indicator(expr.args[1] >= 0) * expr.args[1]
    elif expr.func == Min:
        first = expr.args[0]
        if not isinstance(first, sympy.Number):
            return None
        return Indicator(expr.args[1] <= first) * expr.args[1] + Indicator(expr.args[1] > first) * 2000
    return None

def is_relevant_indicator(expr, var):                
    if expr.func != Indicator:
        return False
    return expr.args[0].has(var)

def prepare(expr):
    if expr.func == Mul:
        result = []
        for arg in expr.args:
            indicator = get_indicator(arg)
            result.append(indicator if indicator != None else arg)
        return Mul(*result)
    return None

def is_atom(cond):
    return cond.func in [Ge, Le, Gt, Lt, Eq, Ne]

def get_inverse(func):
    mapping = {
        Ge: Le,
        Gt: Lt,
        Le: Ge,
        Lt: Gt
    }
    return mapping[func]

def correct(cond, var):
    lhs = cond.lhs
    rhs = cond.rhs
    func = cond.func
    if lhs.func == Add:
        lhs_rel = Add(*[summand for summand in lhs.args if summand.has(var)])
        lhs_irrel = Add(*[summand for summand in lhs.args if not summand.has(var)])
        lhs = lhs_rel
        rhs += -lhs_irrel
    if rhs.func == Add:
        rhs_rel = Add(*[summand for summand in rhs.args if summand.has(var)])
        rhs_irrel = Add(*[summand for summand in rhs.args if not summand.has(var)])
        rhs = rhs_irrel
        lhs += -rhs_rel
    if lhs.func == Mul:
        consts = Mul(*[factor for factor in lhs.args if factor.is_constant()])
        consts_not = Mul(*[factor for factor in lhs.args if not factor.is_constant()])
        lhs = consts_not
        rhs /= consts
        if consts.is_negative:
            func = get_inverse(func)
    if var == lhs:
        return [(func(lhs, rhs), True)]
    if lhs.func != Mul or var not in lhs.args:
        return None
    other_factors = Mul(*[factor for factor in lhs.args if factor != var])
    return [
        (func(var, rhs / other_factors), Gt(other_factors, 0)),
        (func(0, rhs), Eq(other_factors, 0)),
        (get_inverse(func)(var, rhs / other_factors), Lt(other_factors, 0))
    ]
    
def is_linear(cond, var):
    if cond.func == Eq or cond.func == Ne:
        return True
    if var in cond.lhs.free_symbols and var in cond.rhs.free_symbols:
        return False
    return cond.lhs == var or cond.rhs == var

def get_integral(expr, var):
    if expr.func == Add:
        results = [get_integral(arg, var) for arg in expr.args]
        if None in results:
            return None
        return Add(*results)
    elif expr.func == Mul:
        lower = [-oo]
        upper = [oo]
        preamble = []
        result = []
        for arg in expr.args:
            if is_relevant_indicator(arg, var):
                cond = arg.args[0]
                bound = cond.rhs
                if cond.func in [Gt, Ge]:
                    lower.append(bound)
                elif cond.func in [Lt, Le]:
                    upper.append(bound)
            elif arg.func == Indicator:
                preamble.append(arg)
            else:
                result.append(arg)
        lower = Max(*lower)
        upper = Min(*upper)
        preamble.append(Indicator(lower <= upper))
        return Mul(*preamble) * Integral(Mul(*result), (var, lower, upper))
    return None

def linearise(expr, var):
    if expr.func == Mul:
        for factor in expr.args:
            if is_relevant_indicator(factor, var):
                cond = factor.args[0]
                if cond.func == And:
                    other_factors = Mul(*[arg for arg in expr.args if arg != factor])
                    indicators = Mul(*[Indicator(arg) for arg in cond.args])
                    return linearise(indicators * other_factors, var)
                elif is_atom(cond):
                    if not is_linear(cond, var):
                        other_factors = Mul(*[arg for arg in expr.args if arg != factor])
                        conds = correct(cond, var)
                        result = 0
                        for (consequent, antecedent) in conds:
                            if antecedent == True:
                                term = Indicator(consequent) * other_factors
                            else:
                                term = Indicator(antecedent) * Indicator(consequent) * other_factors
                            result += linearise(term, var)
                        return result
                else:
                    return None
        return expr
    return None

def convert_to_bounds(expr, dist: Dist, vars: Set[str]):
    if (isinstance(expr, Integer) or isinstance(expr, Rational)) and expr > 0:
        return True
    elif expr.func == Add:
        return z3.Or([convert_to_bounds(arg, dist, vars) for arg in expr.args])
    elif expr.func == Mul:
        return z3.And([convert_to_bounds(arg, dist, vars) for arg in expr.args])
    elif expr.func == Piecewise and len(expr.args) >= 2:
        size = len(expr.args)
        if expr.args[-1].cond != True:
            raise Exception("Conversion Error")
        args = [convert_to_bounds(pair.expr, dist, vars) for pair in expr.args]
        conds = [convert_equation(pair.cond, dist, vars) for pair in expr.args]
        curr = args[size - 1]
        for j in range(size - 2, -1, -1):
            curr = z3.If(conds[j], args[j], curr)
        return curr
    elif expr.func == DiracDelta:
        expr = convert_expr(expr.args[0], dist, vars)
        return expr == 0 
    raise Exception("Conversion Error")

def fix_indicator(expr):
    arg = Symbol("arg")
    indicator_actual = Lambda(arg, Piecewise((1, arg), (0, True)))
    return expr.replace(Indicator, indicator_actual)

def super_integrate(expr, var):
    if expr.func == Add:
        results = [super_integrate(arg, var) for arg in expr.args]
        if None in results:
            return None
        return Add(*results)

    if expr.has(DiracDelta):
        return delta_integrate(expr, var)
    elif expr.func == Mul:
        expr = prepare(expr)
        expr = linearise(expr, var)
        expr = get_integral(expr, var)
        expr = expr.doit()
        if expr.has(Integral):
            return None
        expr = fix_indicator(expr)
        return expr
    
    return None

def advanced_integrate(expr, vars: Set):
    if expr.func == Add:
        results = [advanced_integrate(arg, vars.copy()) for arg in expr.args]
        if None in results:
            return None
        return Add(*results)
    
    reduced = True
    while reduced and len(vars) != 0:
        reduced = False
        for var in vars:
            expr_cand = super_integrate(expr, var)
            if expr_cand != None:
                vars.remove(var)
                expr = expr_cand
                reduced = True
                break
    if len(vars) != 0:
        return None
    return expr
    